# -*- python -*-
#
# Copyright 2018, 2019, 2020 Liang Chen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''
bedqr.format

file format specification: https://genome.ucsc.edu/FAQ/FAQformat.html
'''

from bedqr.base import FileWithHeaderAndContent


class RowBasedStore(object):
    ROW_TYPE = {
        'HEADER' :  0,
        'BODY'   :  1,
        'EMPTY'  : -1,
        'INVALID': -2,
    }

    def get_cells_from_row(self, row):
        #ret = list()
        ret = row.split(self.FIELD_SEP)
        return ret


class BED(FileWithHeaderAndContent, RowBasedStore):
    '''
    Browser Extensible Data (BED)
    '''

    REQUIRED_FIELDS = ('chrom', 'chromStart', 'chromEnd')
    OPTIONAL_FIELDS = ('name', 'score', 'strand', 'thickStart', 'thickEnd', 'itemRgb', 'blockCount', 'blockSizes', 'blockStarts')
    HEADER_PREFIX = ('browser', 'track')
    FIELD_EMPTY_VALUE = '.'
    FIELD_CHROM_REGEX = ('chr[0-9a-zA-Z_]+', 'scaffold[0-9]+')

    def get_column_names(self, data):
        '''
        :param data: (list of string/tuple)
        '''
        ret = self.REQUIRED_FIELDS + self.OPTIONAL_FIELDS
        detected_column_count = 0 #len(ret)
        if len(data):
            first_row = data[0]
            if isinstance(first_row, (list, tuple)):
                detected_column_count = len(first_row)
            elif isinstance(first_row, (basestring, unicode)):
                #detected_column_count = len(first_row.split(self.FIELD_SEP))
                detected_column_count = len(self.get_cells_from_row(first_row))
            else:
                # TODO: need to find a strategy to deal with unknown/foreign/unpredictable types/classes
                #raise ValueError('cannot parse data row')  # caller or somewhere else deal with this exception
                #detected_column_count = len(first_row)  # call dumb boilerplate instance method
                #detected_column_count = first_row.get_size()  # or the object's public interface
                pass
        return ret[:detected_column_count]

    @classmethod
    def detect_line_type(cls, line):
        CHOICE = cls.ROW_TYPE
        ret = CHOICE['INVALID']

        class FoundMatchSignal(Exception):
            def __init__(self, *args, **kwargs):
                self.found = kwargs.pop('found', None)
                super(FoundMatchSignal, self).__init__(*args, **kwargs)

        try:
            if len(line) == 0:
                raise FoundMatchSignal(found=CHOICE['EMPTY'])

            header_prefix = cls.HEADER_PREFIX
            for item in header_prefix:
                if line.startswith(item):
                    raise FoundMatchSignal(found=CHOICE['HEADER'])

            data_prefix = map(lambda x: x.split('[', 1)[0], cls.FIELD_CHROM_REGEX)
            for item in data_prefix:
                if line.startswith(item):
                    raise FoundMatchSignal(found=CHOICE['BODY'])
        except FoundMatchSignal as match:
            ret = match.found
        #except Exception as ex:
        #    raise
        return ret


class bedDetail(BED):
    '''
    BED detail
    '''
    FIELD_SEP = '\t'

    DETAIL_FIELDS = ('id', 'description')

    def __init__(self, data, **kwargs):
        super(bedDetail, self).__init__(**kwargs)
        if isinstance(data, dict):
            if data.has_key('header') and data.has_key('body'):
                self._parts = list()
                self._parts.append(data['header'])
                self._parts.append(data['body'])

    def get_column_names(self, data):
        columns_a = super(bedDetail, self).get_column_names(data)
        columns_b = ([ None for i in columns_a ] + list(self.DETAIL_FIELDS))[len(self.DETAIL_FIELDS):]
        return map(lambda x, y: y if y is not None else x, columns_a, columns_b)





# -*- python -*-
#
# Copyright 2018, 2019, 2020 Liang Chen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''
bedqr.reader

file I/O wrapper
'''

from bedqr.format import BED, bedDetail


class Reader(object):
    '''
    base-class of file reader
    '''

    def load_file_with_parser(self, parser_cls):
        file_content = self._fp.read()
        lines = file_content.splitlines()
        flags = map(parser_cls.detect_line_type, lines)
        line_count = len(lines)
        parts = {
            'header': [ lines[i] for i in range(line_count) if flags[i]==parser_cls.ROW_TYPE['HEADER'] ],
            'body'  : [ lines[i] for i in range(line_count) if flags[i]==parser_cls.ROW_TYPE['BODY'] ],
        }
        return parser_cls(parts)


class QuickReader(Reader):
    '''
    class should be used by general users
    '''

    def __init__(self, fp, parser=bedDetail):
        self._fp = fp
        self.data = self.load_file_with_parser(parser)





# -*- python -*-
#
# Copyright 2018, 2019, 2020 Liang Chen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
'''
bedqr.base

abstract base class
'''


class MultiPartFile(object):
    '''
    ABC for file format that contains multiple parts
    '''

    def __init__(self, **kwargs):
        super(MultiPartFile, self).__init__()
        # we need to make sure constructing the instance follows the directves in super-class
        # and prevent unwanted side-effect
        self._kwargs = kwargs

    @property
    def parts(self):
        return getattr(self, '_parts')

    def get_part(self, index):
        # make no assumptions about the ordering of the parts here
        _msg = 'sub-class must implement this'
        raise NotImplementedError(_msg)


class FileWithHeaderAndContent(MultiPartFile):
    '''
    file format that has a header part and a content part
    '''
    OFFSET_HEADER = 0
    OFFSET_BODY   = 1

    def get_part(self, index):
        return self.parts[index]

    @property
    def header(self):
        return self.get_part(self.OFFSET_HEADER)

    @property
    def body(self):
        return self.get_part(self.OFFSET_BODY)





#!/usr/bin/env python
# -*- python -*-
#
# Copyright 2018, 2019, 2020 Liang Chen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: Implementation :: CPython',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'Topic :: Software Development :: Libraries :: Python Modules',
]


def find_version(*file_paths):
    import codecs
    import re
    import os

    MSG_VERSION_STRING_NOTFOUND = 'Unable to find version string.'

    def read_src_file(*parts):
        dir_path = os.path.abspath(os.path.dirname(__file__))
        with codecs.open(os.path.join(dir_path, *parts), 'r') as _f:
            return _f.read()

    version_file = read_src_file(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError(MSG_VERSION_STRING_NOTFOUND)


setup(
    name='bedqr',
    version=find_version('bedqr', '__init__.py'),
    packages=['bedqr',],
    author='Liang Chen',
    license='Apache License 2.0',
    platforms='Any',
    keywords=('BED', 'bed file', 'UCSC BED format', 'Bioinformatics'),
    python_requires=">=2.7",
    tests_require=['flake8', 'pytest'],
    classifiers=CLASSIFIERS,
    long_description_markdown_filename='README.md',
)
